#===========================================================================
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#============================================================================
# Copyright (C) 2004, 2005 Mike Wray <mike.wray@hp.com>
# Copyright (C) 2005, 2006 XenSource Ltd
#============================================================================

"""Representation of a single domain.
Includes support for domain construction, using
open-ended configurations.

Author: Mike Wray <mike.wray@hp.com>

"""

import errno
import logging
import string
import time
import threading
import os
import re

import xen.lowlevel.xc
from xen.util import asserts
from xen.util.blkif import blkdev_uname_to_file
from xen.util import security
import balloon
import image
import sxp
import uuid
import XendDomain
import XendRoot
from xen.util.blkif import blkdev_name_to_number
from xen.util.pci import pci_convert_sxp_to_dict

from xen.xend.XendBootloader import bootloader
from xen.xend.XendError import XendError, VmError

from xen.xend.xenstore.xstransact import xstransact, complete
from xen.xend.xenstore.xsutil import GetDomainPath, IntroduceDomain, ResumeDomain
from xen.xend.xenstore.xswatch import xswatch

from xen.xend import arch

"""Shutdown code for poweroff."""
DOMAIN_POWEROFF = 0

"""Shutdown code for reboot."""
DOMAIN_REBOOT   = 1

"""Shutdown code for suspend."""
DOMAIN_SUSPEND  = 2

"""Shutdown code for crash."""
DOMAIN_CRASH    = 3

"""Shutdown code for halt."""
DOMAIN_HALT     = 4

"""Map shutdown codes to strings."""
shutdown_reasons = {
    DOMAIN_POWEROFF: "poweroff",
    DOMAIN_REBOOT  : "reboot",
    DOMAIN_SUSPEND : "suspend",
    DOMAIN_CRASH   : "crash",
    DOMAIN_HALT    : "halt"
    }

reverse_shutdown_reasons = \
    dict([(y, x) for x, y in shutdown_reasons.items()])

HVM_PARAM_CALLBACK_IRQ = 0
HVM_PARAM_STORE_PFN    = 1
HVM_PARAM_STORE_EVTCHN = 2
HVM_PARAM_PAE_ENABLED  = 4
HVM_PARAM_IOREQ_PFN    = 5
HVM_PARAM_BUFIOREQ_PFN = 6
HVM_PARAM_TIMER_MODE   = 10

restart_modes = [
    "restart",
    "destroy",
    "preserve",
    "rename-restart"
    ]

STATE_DOM_OK       = 1
STATE_DOM_SHUTDOWN = 2

SHUTDOWN_TIMEOUT = 30.0
MIGRATE_TIMEOUT = 30.0

ZOMBIE_PREFIX = 'Zombie-'

"""Constants for the different stages of ext. device migration """
DEV_MIGRATE_TEST  = 0
DEV_MIGRATE_STEP1 = 1
DEV_MIGRATE_STEP2 = 2
DEV_MIGRATE_STEP3 = 3

"""Minimum time between domain restarts in seconds."""
MINIMUM_RESTART_TIME = 20

RESTART_IN_PROGRESS = 'xend/restart_in_progress'


xc = xen.lowlevel.xc.xc()
xroot = XendRoot.instance()

log = logging.getLogger("xend.XendDomainInfo")
#log.setLevel(logging.TRACE)


##
# All parameters of VMs that may be configured on-the-fly, or at start-up.
# 
VM_CONFIG_PARAMS = [
    ('name',        str),
    ('on_poweroff', str),
    ('on_reboot',   str),
    ('on_crash',    str),
    ]


##
# Configuration entries that we expect to round-trip -- be read from the
# config file or xc, written to save-files (i.e. through sxpr), and reused as
# config on restart or restore, all without munging.  Some configuration
# entries are munged for backwards compatibility reasons, or because they
# don't come out of xc in the same form as they are specified in the config
# file, so those are handled separately.
ROUNDTRIPPING_CONFIG_ENTRIES = [
    ('uuid',            str),
    ('vcpus',           int),
    ('vcpu_avail',      int),
    ('cpu_cap',         int),
    ('cpu_weight',      float),
    ('memory',          int),
    ('shadow_memory',   int),
    ('maxmem',          int),
    ('bootloader',      str),
    ('bootloader_args', str),
    ('features',        str),
    ('localtime',       int),
    ]

ROUNDTRIPPING_CONFIG_ENTRIES += VM_CONFIG_PARAMS


##
# All entries written to the store.  This is VM_CONFIG_PARAMS, plus those
# entries written to the store that cannot be reconfigured on-the-fly.
#
VM_STORE_ENTRIES = [
    ('uuid',          str),
    ('vcpus',         int),
    ('vcpu_avail',    int),
    ('memory',        int),
    ('shadow_memory', int),
    ('maxmem',        int),
    ('start_time',    float),
    ('bootloader',      str),
    ('bootloader_args', str),
    ]

VM_STORE_ENTRIES += VM_CONFIG_PARAMS


#
# There are a number of CPU-related fields:
#
#   vcpus:       the number of virtual CPUs this domain is configured to use.
#   vcpu_avail:  a bitmap telling the guest domain whether it may use each of
#                its VCPUs.  This is translated to
#                <dompath>/cpu/<id>/availability = {online,offline} for use
#                by the guest domain.
#   cpumap:      a list of bitmaps, one for each VCPU, giving the physical
#                CPUs that that VCPU may use.
#   cpu:         a configuration setting requesting that VCPU 0 is pinned to
#                the specified physical CPU.
#
# vcpus and vcpu_avail settings persist with the VM (i.e. they are persistent
# across save, restore, migrate, and restart).  The other settings are only
# specific to the domain, so are lost when the VM moves.
#


def create(config):
    """Create a VM from a configuration.

    @param config    configuration
    @raise: VmError for invalid configuration
    """

    log.debug("XendDomainInfo.create(%s)", config)

    vm = XendDomainInfo(parseConfig(config))
    try:
        vm.construct()
        vm.initDomain()
        vm.storeVmDetails()
        vm.storeDomDetails()
        vm.registerWatches()
        vm.refreshShutdown()
        return vm
    except:
        log.exception('Domain construction failed')
        vm.destroy()
        raise


def recreate(xeninfo, priv):
    """Create the VM object for an existing domain.  The domain must not
    be dying, as the paths in the store should already have been removed,
    and asking us to recreate them causes problems."""

    log.debug("XendDomainInfo.recreate(%s)", xeninfo)

    assert not xeninfo['dying']

    domid = xeninfo['domid']
    uuid1 = xeninfo['handle']
    xeninfo['uuid'] = uuid.toString(uuid1)
    dompath = GetDomainPath(domid)
    if not dompath:
        raise XendError(
            'No domain path in store for existing domain %d' % domid)

    log.info("Recreating domain %d, UUID %s.", domid, xeninfo['uuid'])
    try:
        vmpath = xstransact.Read(dompath, "vm")
        if not vmpath:
            raise XendError(
                'No vm path in store for existing domain %d' % domid)
        uuid2_str = xstransact.Read(vmpath, "uuid")
        if not uuid2_str:
            raise XendError(
                'No vm/uuid path in store for existing domain %d' % domid)

        uuid2 = uuid.fromString(uuid2_str)

        if uuid1 != uuid2:
            raise XendError(
                'Uuid in store does not match uuid for existing domain %d: '
                '%s != %s' % (domid, uuid2_str, xeninfo['uuid']))

        vm = XendDomainInfo(xeninfo, domid, dompath, True, priv, vmpath = vmpath)

    except Exception, exn:
        if priv:
            log.warn(str(exn))

        vm = XendDomainInfo(xeninfo, domid, dompath, True, priv)
        vm.recreateDom()
        vm.removeVm()
        vm.storeVmDetails()
        vm.storeDomDetails()

    if vm.is_hvm():
        vm.image = image.create(vm,
                                vm.info['image'],
                                vm.info['device'])
        vm.image.recreate()

    vm.registerWatches()
    vm.refreshShutdown(xeninfo)
    return vm


def restore(config, relocating=False):
    """Create a domain and a VM object to do a restore.

    @param config: domain configuration
    """

    log.debug("XendDomainInfo.restore(%s)", config)

    vm = XendDomainInfo(parseConfig(config), None, None, False, False, True)
    vm.relocating = relocating
    try:
        vm.construct()

        try:
            vm._setCPUAffinity()
        except:
            # usually a CPU we want to set affinity to does not exist
            # we just ignore it so that the domain can still be restored
            log.warn("Cannot restore CPU affinity")

        vm.storeVmDetails()
        vm.createChannels()
        vm.createDevices()
        vm.storeDomDetails()
        vm.endRestore()
        vm.relocating = False
        return vm
    except:
        vm.destroy()
        raise


def parseConfig(config):
    def get_cfg(name, conv = None):
        val = sxp.child_value(config, name)

        if conv and not val is None:
            try:
                return conv(val)
            except TypeError, exn:
                raise VmError(
                    'Invalid setting %s = %s in configuration: %s' %
                    (name, val, str(exn)))
        else:
            return val


    log.debug("parseConfig: config is %s", config)

    result = {}

    for e in ROUNDTRIPPING_CONFIG_ENTRIES:
        result[e[0]] = get_cfg(e[0], e[1])

    result['cpu']   = get_cfg('cpu',  int)
    result['cpus']  = get_cfg('cpus')
    result['image'] = get_cfg('image')
    result['timer_mode'] = get_cfg('timer_mode', int)
    tmp_security = get_cfg('security')
    if tmp_security:
        result['security'] = tmp_security

    try:
        if result['image']:
            v = sxp.child_value(result['image'], 'vcpus')
            if result['vcpus'] is None and v is not None:
                result['vcpus'] = int(v)
            elif v is not None and int(v) != result['vcpus']:
                log.warn(('Image VCPUs setting overrides vcpus=%d elsewhere.'
                          '  Using %s VCPUs for VM %s.') %
                         (result['vcpus'], v, result['uuid']))
                result['vcpus'] = int(v)
    except TypeError, exn:
        raise VmError(
            'Invalid configuration setting: vcpus = %s: %s' %
            (sxp.child_value(result['image'], 'vcpus', 1), str(exn)))

    try:
        # support legacy config files with 'cpu' parameter
        # NB: prepending to list to support previous behavior
        #     where 'cpu' parameter pinned VCPU0.
        if result['cpu']:
           if result['cpus']:
               result['cpus'] = "%s,%s" % (str(result['cpu']), result['cpus'])
           else:
               result['cpus'] = str(result['cpu'])

        # Convert 'cpus' to list of ints
        if result['cpus']:
            cpus = []
            if type(result['cpus']) == list:
                # If sxp_cfg was created from config.sxp,
                # the form of 'cpus' is list of string.
                # Convert 'cpus' to list of ints.
                #    ['1']         -> [1]
                #    ['0','2','3'] -> [0,2,3]
                for c in result['cpus']:
                    cpus.append(int(c))
                result['cpus'] = cpus
            else:
                # Convert 'cpus' string to list of ints
                # 'cpus' supports a list of ranges (0-3),
                # separated by commas, and negation, (^1).  
                # Precedence is settled by order of the 
                # string:
                #    "0-3,^1"      -> [0,2,3]
                #    "0-3,^1,1"    -> [0,1,2,3]
                for c in result['cpus'].split(','):
                    if c.find('-') != -1:
                        (x, y) = c.split('-')
                        for i in range(int(x), int(y)+1):
                            cpus.append(int(i))
                    elif c[0] == '^':
                        # remove this element from the list 
                        cpus = [x for x in cpus if x != int(c[1:])]
                    else:
                        cpus.append(int(c))

            result['cpus'] = cpus
        
    except ValueError, exn:
        raise VmError(
            'Invalid configuration setting: cpus = %s: %s' %
            (result['cpus'], exn))

    result['backend'] = []
    for c in sxp.children(config, 'backend'):
        result['backend'].append(sxp.name(sxp.child0(c)))

    result['device'] = []
    for d in sxp.children(config, 'device'):
        c = sxp.child0(d)
        result['device'].append((sxp.name(c), c))

    # Configuration option "restart" is deprecated.  Parse it, but
    # let on_xyz override it if they are present.
    restart = get_cfg('restart')
    if restart:
        def handle_restart(event, val):
            if result[event] is None:
                result[event] = val

        if restart == "onreboot":
            handle_restart('on_poweroff', 'destroy')
            handle_restart('on_reboot',   'restart')
            handle_restart('on_crash',    'destroy')
        elif restart == "always":
            handle_restart('on_poweroff', 'restart')
            handle_restart('on_reboot',   'restart')
            handle_restart('on_crash',    'restart')
        elif restart == "never":
            handle_restart('on_poweroff', 'destroy')
            handle_restart('on_reboot',   'destroy')
            handle_restart('on_crash',    'destroy')
        else:
            log.warn("Ignoring malformed and deprecated config option "
                     "restart = %s", restart)

    result['start_time'] = get_cfg('start_time', float)

    # Normalize UUID wrt to embedded '-' characters
    if result.has_key('uuid') and result['uuid']:
        result['uuid'] = uuid.toString(uuid.fromString(result['uuid']))

    log.debug("parseConfig: result is %s", result)
    return result


def domain_by_name(name):
    return XendDomain.instance().domain_lookup_by_name_nr(name)


def shutdown_reason(code):
    """Get a shutdown reason from a code.

    @param code: shutdown code
    @type  code: int
    @return: shutdown reason
    @rtype:  string
    """
    return shutdown_reasons.get(code, "?")

def dom_get(dom):
    """Get info from xen for an existing domain.

    @param dom: domain id
    @return: info or None
    """
    try:
        domlist = xc.domain_getinfo(dom, 1)
        if domlist and dom == domlist[0]['domid']:
            return domlist[0]
    except Exception, err:
        # ignore missing domain
        log.trace("domain_getinfo(%d) failed, ignoring: %s", dom, str(err))
    return None

from xen.xend.server.pciif import parse_pci_name, PciDevice,\
    get_assigned_pci_devices, get_all_assigned_pci_devices

def do_FLR(domid, is_hvm):
    dev_str_list = get_assigned_pci_devices(domid)

    for dev_str in dev_str_list:
        (dom, b, d, f) = parse_pci_name(dev_str)
        try:
            dev = PciDevice(dom, b, d, f)
        except Exception, e:
            raise VmError("pci: failed to locate device and "+
                    "parse it's resources - "+str(e))
        dev.do_FLR(is_hvm, xroot.get_pci_strict_checking())

class XendDomainInfo:

    def __init__(self, info, domid = None, dompath = None, augment = False,
                 priv = False, resume = False, vmpath = None):

        self.info = info
        self.relocating = False

        if not self.infoIsSet('uuid'):
            self.info['uuid'] = uuid.toString(uuid.create())

        if domid is not None:
            self.domid = domid
        elif 'domid' in info:
            self.domid = int(info['domid'])
        else:
            self.domid = None

        # Find a unique /vm/<uuid>/<integer> path if not specified.
        # This avoids conflict between pre-/post-migrate domains when doing
        # localhost relocation.
        self.vmpath = vmpath
        i = 0
        while self.vmpath == None:
            self.vmpath = XendDomain.VMROOT + self.info['uuid']
            if i != 0:
                self.vmpath = self.vmpath + '-' + str(i)

            try:
                if self.readVm("uuid"):
                    self.vmpath = None
                i = i + 1
            except:
                pass

        self.dompath = dompath

        if augment:
            self.augmentInfo(priv)

        self.validateInfo()

        self.image = None
        self.security = None
        self.store_port = None
        self.store_mfn = None
        self.console_port = None
        self.console_mfn = None

        self.native_protocol = None

        self.vmWatch = None
        self.shutdownWatch = None
        self.clean = False
        self.restart_in_progress = False

        self.shutdownStartTime = None
        
        self.state = STATE_DOM_OK
        self.state_updated = threading.Condition()
        self.refresh_shutdown_lock = threading.Condition()

        self.setResume(resume)

    ## private:

    def device_tuple_value_from_dev_info(self, dev_info, key):
         for x in dev_info:
             if (type(x) != str):
                 for xx in x:
                     if (xx[0] == key):
                         return xx[1]
         return None

    # This function translates all block device modes (incl. aliases) to
    # one common label per each device mode. Those modes can be:
    #  read-only (ro), write-exclusive (wx) and write-shared (ws)
    def block_device_mode_translate(self, mode):
         # Device modes can be read-only (ro), write-exclusive (wx) or
         # write-shared (ws), otherwise an error is raised
         if mode == "w" or mode == "wr":
             return "wx"
         elif mode == "r" or mode == "ro":
             return "ro"
         elif mode == "!" or mode == "w!":
             return "ws"

         # If no mode defined we consider this as write-exclusive
         return "wx"

    # Detect device duplicates for vbd, tap and vif devices for domain and
    # duplicate unames in global context not to destroy virtual block devices
    def device_duplicate_check(self, dev_type, dev_config):
         # Enumerate all devices for all domains
         allSxprs = []
         val = XendDomain.instance().domains.values()
         for v in val:
             sxpr = v.getDeviceSxprs(dev_type)
             for key in sxpr:
                try:
                    index = allSxprs.index(key)
                except:
                    allSxprs.append(key)

         # Enumerate devices for current domain
         sxpr = self.getDeviceSxprs(dev_type)

         # For vif interface we won't check cross-domain
         if sxpr == None and dev_type == 'vif':
            return

         # Preset None values to all variables we'll be checking
         new_uname = None
         uname = None
         dev = None
         mac = None
         mode = None

         # Disk device
         if dev_type in ['vbd', 'tap']:
             for x in dev_config:
                 if type(x) != str and (x[0] in ['uname', 'dev', 'mode']):
                     if x[0] == 'uname':
                         new_uname = x[1]
                     if x[0] == 'dev':
                         dev = x[1]
                     if x[0] == 'mode':
                         mode = x[1]

             # If we don't have uname entry (happens in virt-manager) return
             if new_uname == None:
                 return

             uname = new_uname.split(":")[len(new_uname.split(":"))-1]
             # We need to allow when uname is zero length, eg. hdc:cdrom device
             if len(uname) == 0:
                 log.debug("Null uname when attaching disk device, allowing %s..."
                           % dev)
                 return

             log.debug("Checking for duplicate for uname: %s [%s], dev: %s, mode: %s"
                       % (uname, new_uname, dev, mode))
             # No device in dev found
             if dev == None:
                 return

             devid = blkdev_name_to_number(dev)
             if devid == None:
                 return

             for o_dev_info in sxpr:
                 backend = self.device_tuple_value_from_dev_info(o_dev_info, "backend")
                 params = xstransact.Read(backend, "params")
                 dev_uname = params.split(":")[len(params.split(":"))-1]
                 if uname == dev_uname:
                     raise VmError('The uname "%s" is already defined' %
                                   dev_uname)
                 blkdev = xstransact.Read(backend, "dev")
                 if blkdev is None:
                     return

                 blkdevid = blkdev_name_to_number(blkdev)
                 if blkdevid != None and devid == blkdevid:
                     raise VmError('The device "%s" is already defined' %
                                   blkdev)

             tMode = self.block_device_mode_translate(mode)

             # Device/uname not found in the context of current domain but we
             # need to have a look to global context. We deny addition of device
             # in those cases:
             #   1. We're adding read-only disk that's already used as write-exclusive
             #   2. We're adding write-shared disk that's already used as write-exclusive
             #   3. We're adding write-exclusive disk that's already used
             #   4. We're adding read-only disk that's already used as write-shared
             for o_dev_info in allSxprs:
                 backend = self.device_tuple_value_from_dev_info(o_dev_info, "backend")
                 params = xstransact.Read(backend, "params")
                 aMode = self.block_device_mode_translate(
                                     xstransact.Read(backend, "mode") )
                 dev_uname = params.split(":")[len(params.split(":"))-1]
                 if uname == dev_uname:
                     if ((tMode == "ro" and aMode == "wx")
                       or (tMode == "ws" and aMode == "wx")
                       or (tMode == "ro" and aMode == "ws")
                       or (tMode == "wx")):
                         raise VmError('The uname "%s" is already used by another domain' %
                                       dev_uname)

         # Virtual network adapter
         elif dev_type == 'vif':
             for x in dev_config:
                 if type(x) != str and (x[0] == 'mac'):
                     mac = x[1]

             for o_dev_info in sxpr:
                 o_mac = self.device_tuple_value_from_dev_info(o_dev_info, "mac")
                 if mac == o_mac:
                     raise VmError('The mac "%s" is already defined' %
                                   mac)


    def readVMDetails(self, params):
        """Read the specified parameters from the store.
        """
        try:
            return self.gatherVm(*params)
        except ValueError:
            # One of the int/float entries in params has a corresponding store
            # entry that is invalid.  We recover, because older versions of
            # Xend may have put the entry there (memory/target, for example),
            # but this is in general a bad situation to have reached.
            log.exception(
                "Store corrupted at %s!  Domain %d's configuration may be "
                "affected.", self.vmpath, self.domid)
            return []

    def  _getDeviceInfo_pci(self):
        for dev_type, dev_info in self.info['device']:
            if dev_type != 'pci':
                continue
            return dev_info
        return None

    def storeChanged(self, _):
        log.trace("XendDomainInfo.storeChanged");

        changed = False
        
        def f(x, y):
            if y is not None and self.info[x[0]] != y:
                self.info[x[0]] = y
                changed = True

        map(f, VM_CONFIG_PARAMS, self.readVMDetails(VM_CONFIG_PARAMS))

        im = self.readVm('image')
        current_im = self.info['image']
        if (im is not None and
            (current_im is None or sxp.to_string(current_im) != im)):
            self.info['image'] = sxp.from_string(im)
            changed = True

        if changed:
            # Update the domain section of the store, as this contains some
            # parameters derived from the VM configuration.
            self.storeDomDetails()

        return 1


    def augmentInfo(self, priv):
        """Augment self.info, as given to us through {@link #recreate}, with
        values taken from the store.  This recovers those values known to xend
        but not to the hypervisor.
        """
        def useIfNeeded(name, val):
            if not self.infoIsSet(name) and val is not None:
                self.info[name] = val

        if priv:
            entries = VM_STORE_ENTRIES[:]
            entries.remove(('memory', int))
            entries.remove(('maxmem', int))
        else:
            entries = VM_STORE_ENTRIES
        entries.append(('image', str))
        entries.append(('security', str))

        map(lambda x, y: useIfNeeded(x[0], y), entries,
            self.readVMDetails(entries))

        device = []
        disks_done = 0
        for c in controllerClasses:
            if c in ('vbd', 'tap'):
                if not disks_done:
                    disks_done = 1
                    devconfig = self.getDiskConfigurations()
                    device.extend(map(lambda x: (x[0], x), devconfig))
            else:
                devconfig = self.getDeviceConfigurations(c)
                if devconfig:
                    device.extend(map(lambda x: (c, x), devconfig))
        useIfNeeded('device', device)


    def validateInfo(self):
        """Validate and normalise the info block.  This has either been parsed
        by parseConfig, or received from xc through recreate and augmented by
        the current store contents.
        """
        def defaultInfo(name, val):
            if not self.infoIsSet(name):
                self.info[name] = val()

        try:
            defaultInfo('name',         lambda: "Domain-%d" % self.domid)
            defaultInfo('on_poweroff',  lambda: "destroy")
            defaultInfo('on_reboot',    lambda: "restart")
            defaultInfo('on_crash',     lambda: "restart")
            defaultInfo('features',     lambda: "")
            defaultInfo('cpu',          lambda: None)
            defaultInfo('cpus',         lambda: [])
            defaultInfo('cpu_cap',      lambda: 0)
            defaultInfo('cpu_weight',   lambda: 256)

            # some domains don't have a config file (e.g. dom0 )
            # to set number of vcpus so we derive available cpus
            # from max_vcpu_id which is present for running domains.
            if not self.infoIsSet('vcpus') and self.infoIsSet('max_vcpu_id'):
                avail = int(self.info['max_vcpu_id'])+1
            else:
                avail = int(1)

            defaultInfo('vcpus',        lambda: avail)
            defaultInfo('online_vcpus', lambda: self.info['vcpus'])
            defaultInfo('max_vcpu_id',  lambda: self.info['vcpus']-1)
            defaultInfo('vcpu_avail',   lambda: (1 << self.info['vcpus']) - 1)

            defaultInfo('memory',       lambda: 0)
            defaultInfo('shadow_memory', lambda: 0)
            defaultInfo('maxmem',       lambda: 0)
            defaultInfo('bootloader',   lambda: None)
            defaultInfo('bootloader_args', lambda: None)            
            defaultInfo('backend',      lambda: [])
            defaultInfo('device',       lambda: [])
            defaultInfo('image',        lambda: None)
            defaultInfo('security',     lambda: None)

            self.check_name(self.info['name'])

            if isinstance(self.info['image'], str):
                self.info['image'] = sxp.from_string(self.info['image'])

            if isinstance(self.info['security'], str):
                self.info['security'] = sxp.from_string(self.info['security'])

            if self.info['memory'] == 0:
                if self.infoIsSet('mem_kb'):
                    self.info['memory'] = (self.info['mem_kb'] + 1023) / 1024
            if self.info['memory'] <= 0:
                raise VmError('Invalid memory size')

            if self.info['maxmem'] < self.info['memory']:
                self.info['maxmem'] = self.info['memory']

            for (n, c) in self.info['device']:
                if not n or not c or n not in controllerClasses:
                    raise VmError('invalid device (%s, %s)' %
                                  (str(n), str(c)))

            for event in ['on_poweroff', 'on_reboot', 'on_crash']:
                if self.info[event] not in restart_modes:
                    raise VmError('invalid restart event: %s = %s' %
                                  (event, str(self.info[event])))

        except KeyError, exn:
            log.exception(exn)
            raise VmError('Unspecified domain detail: %s' % exn)


    def readVm(self, *args):
        return xstransact.Read(self.vmpath, *args)

    def writeVm(self, *args):
        return xstransact.Write(self.vmpath, *args)

    def removeVm(self, *args):
        return xstransact.Remove(self.vmpath, *args)

    def gatherVm(self, *args):
        return xstransact.Gather(self.vmpath, *args)

    def listRecursiveVm(self, *args):
        return xstransact.ListRecursive(self.vmpath, *args)
    

    ## public:

    def storeVm(self, *args):
        return xstransact.Store(self.vmpath, *args)


    ## private:

    def readDom(self, *args):
        return xstransact.Read(self.dompath, *args)

    def writeDom(self, *args):
        return xstransact.Write(self.dompath, *args)

    def gatherDom(self, *args):
        return xstransact.Gather(self.dompath, *args)

    ## public:

    def removeDom(self, *args):
        return xstransact.Remove(self.dompath, *args)

    def recreateDom(self):
        complete(self.dompath, lambda t: self._recreateDom(t))

    def _recreateDom(self, t):
        t.remove()
        t.mkdir()
        #
        # The first permission record on any node indicates
        # the domain owner. Thus any permissions associated
        # with the first record are ignored, since the owner
        # is allowed todo anything.
        #
        # Thus to make the node read-only to the guest
        # we must explicitly give Dom0 permission making
        # it the owner, even though Dom0 already has all the
        # permissions inherited from the parent.
        #
        t.set_permissions({'dom': 0 }, { 'dom' : self.domid, 'read' : True })
        t.write('vm', self.vmpath)
        for i in [ 'device', 'control', 'error' ]:
            t.mkdir(i)
            t.set_permissions(i, { 'dom' : self.domid })


    ## private:

    def storeDom(self, *args):
        return xstransact.Store(self.dompath, *args)


    ## public:

    def completeRestore(self, store_mfn, console_mfn):

        log.debug("XendDomainInfo.completeRestore")

        self.store_mfn = store_mfn
        self.console_mfn = console_mfn

        self.introduceDomain()
        if self.is_hvm():
            self.image = image.create(self,
                                      self.info['image'],
                                      self.info['device'])
            if self.image:
                self.image.createDeviceModel(True)
        self.storeDomDetails()
        self.registerWatches()
        self.refreshShutdown()

        log.debug("XendDomainInfo.completeRestore done")


    def testvifsComplete(self):
        """ In case vifs are released and then created for the same
        domain, we need to wait the device shut down.
        """
        start = time.time()
        while True:
            test = 0
            diff = time.time() - start
            for i in self.getDeviceController('vif').deviceIDs():
                test = 1
                log.info("Dev %s still active, looping...", i)
                time.sleep(0.1)

            if test == 0:
                break
            if diff >= MIGRATE_TIMEOUT:
                log.info("Dev still active but hit max loop timeout")
                break

    def storeVmDetails(self):
        to_store = {}

        for k in VM_STORE_ENTRIES:
            if self.infoIsSet(k[0]):
                to_store[k[0]] = str(self.info[k[0]])

        if self.infoIsSet('image'):
            to_store['image'] = sxp.to_string(self.info['image'])

        if self.infoIsSet('security'):
            security = self.info['security']
            to_store['security'] = sxp.to_string(security)
            for idx in range(0, len(security)):
                if security[idx][0] == 'access_control':
                    to_store['security/access_control'] = sxp.to_string([ security[idx][1] , security[idx][2] ])
                    for aidx in range(1, len(security[idx])):
                        if security[idx][aidx][0] == 'label':
                            to_store['security/access_control/label'] = security[idx][aidx][1]
                        if security[idx][aidx][0] == 'policy':
                            to_store['security/access_control/policy'] = security[idx][aidx][1]
                if security[idx][0] == 'ssidref':
                    to_store['security/ssidref'] = str(security[idx][1])

        if not self.readVm('xend/restart_count'):
            to_store['xend/restart_count'] = str(0)

        log.debug("Storing VM details: %s", to_store)

        self.writeVm(to_store)
        self.setVmPermissions()


    def setVmPermissions(self):
        """Allow the guest domain to read its UUID.  We don't allow it to
        access any other entry, for security."""
        xstransact.SetPermissions('%s/uuid' % self.vmpath,
                                  { 'dom' : self.domid,
                                    'read' : True,
                                    'write' : False })


    def storeDomDetails(self):
        to_store = {
            'domid':              str(self.domid),
            'vm':                 self.vmpath,
            'name':               self.info['name'],
            'console/limit':      str(xroot.get_console_limit() * 1024),
            'memory/target':      str(self.info['memory'] * 1024)
            }

        def f(n, v):
            if v is not None:
                to_store[n] = str(v)

        f('console/port',     self.console_port)
        f('console/ring-ref', self.console_mfn)
        f('store/port',       self.store_port)
        f('store/ring-ref',   self.store_mfn)

        to_store.update(self.vcpuDomDetails())

        log.debug("Storing domain details: %s", to_store)

        self.writeDom(to_store)


    ## private:

    def vcpuDomDetails(self):
        def availability(n):
            if self.info['vcpu_avail'] & (1 << n):
                return 'online'
            else:
                return 'offline'

        result = {}
        for v in range(0, self.info['vcpus']):
            result["cpu/%d/availability" % v] = availability(v)
        return result


    ## public:

    def registerWatches(self):
        """Register a watch on this VM's entries in the store, and the
        domain's control/shutdown node, so that when they are changed
        externally, we keep up to date.  This should only be called by {@link
        #create}, {@link #recreate}, or {@link #restore}, once the domain's
        details have been written, but before the new instance is returned."""
        self.vmWatch = xswatch(self.vmpath, self.storeChanged)
        self.shutdownWatch = xswatch(self.dompath + '/control/shutdown',
                                     self.handleShutdownWatch)


    def getDomid(self):
        return self.domid

    def setName(self, name, to_store = True):
        self.check_name(name)
        self.info['name'] = name
        if to_store:
            self.storeVm("name", name)

    def getName(self):
        return self.info['name']

    def getDomainPath(self):
        return self.dompath


    def getStorePort(self):
        """For use only by image.py and XendCheckpoint.py."""
        return self.store_port


    def getConsolePort(self):
        """For use only by image.py and XendCheckpoint.py"""
        return self.console_port

    def getFeatures(self):
        """For use only by image.py."""
        return self.info['features']

    def getVCpuCount(self):
        return self.info['vcpus']


    def setVCpuCount(self, vcpus):
        self.info['vcpu_avail'] = (1 << vcpus) - 1
        self.storeVm('vcpu_avail', self.info['vcpu_avail'])
        self.writeDom(self.vcpuDomDetails())

    def getCap(self):
        return self.info['cpu_cap']

    def setCap(self, cpu_cap):
        self.info['cpu_cap'] = cpu_cap

    def getWeight(self):
        # Compatibility fix to allow restore/migration from older version of user-space tools
        x = self.info['cpu_weight']
        if (type(x) == type(float)):
            x = 256
            log.debug("Float value for cpu_weight detected, reset to default value of 256")
        return int(x)

    def setWeight(self, cpu_weight):
        self.info['cpu_weight'] = cpu_weight      

    def getLabel(self):
        return security.get_security_info(self.info, 'label')

    def getMemoryTarget(self):
        """Get this domain's target memory size, in KB."""
        return self.info['memory'] * 1024

    def getMemoryMaximum(self):
        """Get this domain's maxmimum memory size, in KB."""
        return self.info['maxmem'] * 1024

    def getResume(self):
        return "%s" % self.info['resume']

    def endRestore(self):
        self.setResume(False)

    def setResume(self, state):
        self.info['resume'] = state

    def getRestartCount(self):
        return self.readVm('xend/restart_count')

    def refreshShutdown(self, xeninfo = None):
        # If set at the end of this method, a restart is required, with the
        # given reason.  This restart has to be done out of the scope of
        # refresh_shutdown_lock.
        restart_reason = None
        
        self.refresh_shutdown_lock.acquire()
        try:
            if xeninfo is None:
                xeninfo = dom_get(self.domid)
                if xeninfo is None:
                    # The domain no longer exists.  This will occur if we have
                    # scheduled a timer to check for shutdown timeouts and the
                    # shutdown succeeded.  It will also occur if someone
                    # destroys a domain beneath us.  We clean up the domain,
                    # just in case, but we can't clean up the VM, because that
                    # VM may have migrated to a different domain on this
                    # machine.
                    self.cleanupDomain()
                    return

            if xeninfo['dying']:
                # Dying means that a domain has been destroyed, but has not
                # yet been cleaned up by Xen.  This state could persist
                # indefinitely if, for example, another domain has some of its
                # pages mapped.  We might like to diagnose this problem in the
                # future, but for now all we do is make sure that it's not us
                # holding the pages, by calling cleanupDomain.  We can't
                # clean up the VM, as above.
                self.cleanupDomain()
                return

            elif xeninfo['crashed']:
                if self.readDom('xend/shutdown_completed'):
                    # We've seen this shutdown already, but we are preserving
                    # the domain for debugging.  Leave it alone.
                    return

                log.warn('Domain has crashed: name=%s id=%d.',
                         self.info['name'], self.domid)

                if xroot.get_enable_dump():
                    try:
                        if self.readDom('xend/crash_dump'):
                            log.info("Skipping duplicated automatic crash dump")
                        else:
                            log.info("Starting automatic crash dump")
                            self.writeDom('xend/crash_dump', 'True')
                            self.dumpCore()
                    except:
                        # This error has been logged, nothing more we can
                        # do in this context & don't want to break the
                        # xm list command by throwing an exception every
                        # time it is run
                        pass

                restart_reason = 'crash'

            elif xeninfo['shutdown']:
                if self.readDom('xend/shutdown_completed'):
                    # We've seen this shutdown already, but we are preserving
                    # the domain for debugging.  Leave it alone.
                    return

                else:
                    reason = shutdown_reason(xeninfo['shutdown_reason'])

                    log.info('Domain has shutdown: name=%s id=%d reason=%s.',
                             self.info['name'], self.domid, reason)

                    self.clearRestart()

                    if reason == 'suspend':
                        self.state_set(STATE_DOM_SHUTDOWN)
                        # Don't destroy the domain.  XendCheckpoint will do
                        # this once it has finished.  However, stop watching
                        # the VM path now, otherwise we will end up with one
                        # watch for the old domain, and one for the new.
                        self.unwatchVm()
                    elif reason in ['poweroff', 'reboot']:
                        restart_reason = reason
                    else:
                        self.destroy()

            elif self.dompath is None:
                # We have yet to manage to call introduceDomain on this
                # domain.  This can happen if a restore is in progress, or has
                # failed.  Ignore this domain.
                pass
            else:
                # Domain is alive.  If we are shutting it down, then check
                # the timeout on that, and log a message if unresponsive

                if self.shutdownStartTime:
                    timeout = (SHUTDOWN_TIMEOUT - time.time() +
                               self.shutdownStartTime)
                    if (timeout < 0 and not self.readDom('xend/unresponsive')):
                        log.info(
                            "Domain shutdown timeout expired: name=%s id=%s",
                            self.info['name'], self.domid)
                        self.storeDom('xend/unresponsive', 'True')
        finally:
            self.refresh_shutdown_lock.release()

        if restart_reason and not self.restart_in_progress:
            self.restart_in_progress = True
            threading.Thread(target = self.maybeRestart,
                             args = (restart_reason,)).start()


    def handleShutdownWatch(self, _):
        log.debug('XendDomainInfo.handleShutdownWatch')
        
        reason = self.readDom('control/shutdown')

        if reason and reason != 'suspend':
            sst = self.readDom('xend/shutdown_start_time')
            now = time.time()
            if sst:
                self.shutdownStartTime = float(sst)
                timeout = float(sst) + SHUTDOWN_TIMEOUT - now
            else:
                self.shutdownStartTime = now
                self.storeDom('xend/shutdown_start_time', now)
                timeout = SHUTDOWN_TIMEOUT

            log.trace(
                "Scheduling refreshShutdown on domain %d in %ds.",
                self.domid, timeout)
            threading.Timer(timeout, self.refreshShutdown).start()

        return True


    def shutdown(self, reason):
        if not reason in shutdown_reasons.values():
            raise XendError('Invalid reason: %s' % reason)
        if self.domid == 0:
            raise XendError("Can't specify Domain-0")

        # If domain is paused, we can't shutdown the domain
        if self.info['paused'] == 1:
            raise VmError("Can't shutdown/save the domain since the domain is paused;"
                                        "unpause it first if you want to shutdown/save")

        self.storeDom("control/shutdown", reason)

        # HVM domain shuts itself down only if it has PV drivers
        if self.is_hvm():
            hvm_pvdrv = xc.hvm_get_param(self.domid, HVM_PARAM_CALLBACK_IRQ)
            if not hvm_pvdrv:
                code = reverse_shutdown_reasons[reason]
                xc.domain_destroy_hook(self.domid)
                log.info("HVM save:remote shutdown dom %d!", self.domid)
                xc.domain_shutdown(self.domid, code)
            

    ## private:

    def clearRestart(self):
        self.removeDom("xend/shutdown_start_time")


    def maybeRestart(self, reason):
        # Dispatch to the correct method based upon the configured on_{reason}
        # behaviour.
        {"destroy"        : self.destroy,
         "restart"        : self.restart,
         "preserve"       : self.preserve,
         "rename-restart" : self.renameRestart}[self.info['on_' + reason]]()


    def renameRestart(self):
        self.restart(True)


    def dumpCore(self,corefile=None):
        """Create a core dump for this domain.  Nothrow guarantee."""
        
        try:
            if not corefile:
                this_time = time.strftime("%Y-%m%d-%H%M.%S", time.localtime())
                corefile = "/var/lib/xen/dump/%s-%s.%s.core" % (this_time,
                                  self.info['name'], self.domid)
                
            if os.path.isdir(corefile):
                raise XendError("Cannot dump core in a directory: %s" %
                                corefile)
            
            xc.domain_dumpcore(self.domid, corefile)
        except RuntimeError, ex:
            corefile_incomp = corefile+'-incomplete'
            os.rename(corefile, corefile_incomp)
            log.exception("XendDomainInfo.dumpCore failed: id = %s name = %s",
                          self.domid, self.info['name'])
            raise XendError("Failed to dump core: %s" %  str(ex))

    ## public:

    def setMemoryTarget(self, target):
        """Set the memory target of this domain.
        @param target In MiB.
        """
        if target <= 0:
            raise XendError('Invalid memory size')
        
        log.debug("Setting memory target of domain %s (%d) to %d MiB.",
                  self.info['name'], self.domid, target)
        
        self.info['memory'] = target
        self.storeVm("memory", target)
        self.storeDom("memory/target", target << 10)


    def update(self, info = None):
        """Update with info from xc.domain_getinfo().
        """

        log.trace("XendDomainInfo.update(%s) on domain %d", info, self.domid)
        if not info:
            info = dom_get(self.domid)
            if not info:
                return
            
        #manually update ssidref / security fields
        if security.on() and info.has_key('ssidref'):
            if (info['ssidref'] != 0) and self.info.has_key('security'):
                security_field = self.info['security']
                if not security_field:
                    #create new security element
                    self.info.update({'security': [['ssidref', str(info['ssidref'])]]})
            #ssidref field not used any longer
        info.pop('ssidref')

        self.info.update(info)
        self.validateInfo()
        self.refreshShutdown(info)

        log.trace("XendDomainInfo.update done on domain %d: %s", self.domid,
                  self.info)

    def getSysMem(self):
        info = dom_get(self.domid)
        xs_memory=int(info['mem_kb'])/1024
        return xs_memory

    ## private:

    def state_set(self, state):
        self.state_updated.acquire()
        try:
            if self.state != state:
                self.state = state
                self.state_updated.notifyAll()
        finally:
            self.state_updated.release()


    ## public:

    def waitForShutdown(self):
        self.state_updated.acquire()
        try:
            while self.state == STATE_DOM_OK:
                self.state_updated.wait()
        finally:
            self.state_updated.release()


    def __str__(self):
        s = "<domain"
        s += " id=" + str(self.domid)
        s += " name=" + self.info['name']
        s += " memory=" + str(self.info['memory'])
        s += ">"
        return s

    __repr__ = __str__


    ## private:

    def createDevice(self, deviceClass, devconfig):
        if not self.relocating:
            self.device_duplicate_check(deviceClass, devconfig)

        return self.getDeviceController(deviceClass).createDevice(devconfig)


    def waitForDevices_(self, deviceClass):
        return self.getDeviceController(deviceClass).waitForDevices()


    def waitForDevice(self, deviceClass, devid):
        return self.getDeviceController(deviceClass).waitForDevice(devid)


    def reconfigureDevice(self, deviceClass, devid, devconfig):
        return self.getDeviceController(deviceClass).reconfigureDevice(
            devid, devconfig)


    ## public:

    def getPCIConfList(self):
        dev_info = self._getDeviceInfo_pci()
        if dev_info is None:
            return []

        dev_list = sxp.children(dev_info, 'dev')
        pci_conf_list = []
        for x in dev_list:
            seg  = int(sxp.child_value(x, "domain"), 16)
            bus  = int(sxp.child_value(x, "bus"), 16)
            dev  = int(sxp.child_value(x, "slot"), 16)
            func = int(sxp.child_value(x, "func"), 16)
            vslt = sxp.child_value(x, "vslt")
            if vslt is not None:
                vslt = int(vslt, 16)

            pci_dict = {
                'domain': seg,
                'bus':    bus,
                'slot':   dev,
                'func':   func,
                'vslt':   vslt
                }
            pci_conf_list.append(pci_dict)

        return pci_conf_list

    def destroyPCIDevice(self, dev_sxp):
        log.debug("destroyPCIDevice called %s", dev_sxp)

        if not self.is_hvm():
            raise VmError("only HVM guest support pci detach")

        #all the PCI devs share one conf node
        devid = '0'

        dev_type = sxp.name(dev_sxp)
        dev_config = pci_convert_sxp_to_dict(dev_sxp)
        new_pci = dev_config['devs'][0]

        new_seg  = int(new_pci['domain'], 16)
        new_bus  = int(new_pci['bus'], 16)
        new_dev  = int(new_pci['slot'], 16)
        new_func = int(new_pci['func'], 16)

        dev_info = self._getDeviceInfo_pci()
        if dev_info is None:
            raise VmError("No pci pass-through device")

        dev_list = sxp.children(dev_info, 'dev')
        num_devs = len(dev_list)

        devnum = -1
        vslot = None
        for x in dev_info:
            devnum += 1
            if sxp.name(x) != 'dev':
                continue
            seg  = int(sxp.child_value(x, "domain"), 16)
            bus  = int(sxp.child_value(x, "bus"), 16)
            dev  = int(sxp.child_value(x, "slot"), 16)
            func = int(sxp.child_value(x, "func"), 16)

            if (seg  == new_seg and
               bus  == new_bus and
               dev  == new_dev and
               func == new_func ):
                vslot = sxp.child_value(x, "vslt")
                break

        if vslot is None:
            raise VmError("device doesn't exist")

        # Check the co-assignment.
        # To pci-detach a device D from domN, we should ensure: for each DD in the
        # list of D's co-assignment devices, DD is not assigned (to domN).
        #
        from xen.xend.server.pciif import PciDevice
        try:
            pci_device = PciDevice(new_seg, new_bus, new_dev, new_func)
        except Exception, e:
            raise VmError("pci: failed to locate device and "+
                    "parse it's resources - "+str(e))
        coassignment_list = pci_device.find_coassigned_devices()
        coassignment_list.remove(pci_device.name)
        assigned_pci_device_str_list = get_assigned_pci_devices(self.domid)
        for pci_str in coassignment_list:
            if pci_str in assigned_pci_device_str_list:
                raise VmError(("pci: failed to pci-detach %s from dom%d" + \
                    " because one of its co-assignment device %s is still " + \
                    " assigned to the domain." \
                    )% (pci_device.name, self.domid, pci_str))

        bdf_str = "%s:%s:%s.%s" % (new_pci['domain'],
                new_pci['bus'],
                new_pci['slot'],
                new_pci['func'])
        log.info("destroyPCIDevice:%s:%s!", x, bdf_str)
        self.image.signalDeviceModel('pci-rem', 'pci-removed', bdf_str)

        if num_devs > 1:
            del dev_info[devnum]
            self.reconfigureDevice('pci', devid, dev_info)
        else:
            self.getDeviceController('pci').destroyDevice(devid, True)
            for type, info in self.info['device']:
                if type != 'pci':
                    continue
                delnum = self.info['device'].index((type, info))
                self.info['device'].remove((type, info))

        from xen.xend.server.pciif import PciDevice
        try:
            device = PciDevice(new_seg, new_bus, new_dev, new_func)
        except Exception, e:
            raise VmError("pci: failed to locate device and "+
                "parse it's resources - "+str(e))
        device.do_FLR(self.is_hvm(), xroot.get_pci_strict_checking())

        return 0

    def destroyDevice(self, deviceClass, devid, force=None):
        if deviceClass == 'pci':
            return self.destroyPCIDevice(devid)

        if type(devid) is str:
            devicePath = '%s/device/%s' % (self.dompath, deviceClass)
            for entry in xstransact.List(devicePath):
                backend = xstransact.Read('%s/%s' % (devicePath, entry),
                                          "backend")
                devName = xstransact.Read(backend, "dev")
                if devName == devid:
                    # We found the integer matching our devid, use it instead
                    devid = entry
                    break

        # Work around broken clients passing 'vbd' instead of 'tap'
        if deviceClass == "vbd":
            altPath = "%s/device/tap/%d/frontend" % (self.vmpath, int(devid))
            log.debug("Checkingif 'vbd' device %d is really a'tap' device" %int(devid))
            if xstransact.Read(altPath):
                log.debug("Converting 'vbd' device to 'tap'")
                deviceClass = "tap"

        return self.getDeviceController(deviceClass).destroyDevice(devid, force)


    def getDeviceSxprs(self, deviceClass):
        if deviceClass == 'pci':
            return self.getPCIConfList()
        return self.getDeviceController(deviceClass).sxprs()


    ## private:

    def getDeviceConfigurations(self, deviceClass):
        return self.getDeviceController(deviceClass).configurations()


    def _releaseDevices(self, suspend = False):
        """Release all domain's devices.  Nothrow guarantee."""
        if suspend and self.image:
            self.image.destroy(suspend)
            return

        while True:
            t = xstransact("%s/device" % self.dompath)
            for devclass in controllerClasses.keys():
                for dev in t.list(devclass):
                    try:
                        t.remove(dev)
                    except:
                        # Log and swallow any exceptions in removal --
                        # there's nothing more we can do.
                        log.exception(
                           "Device release failed: %s; %s; %s",
                           self.info['name'], devclass, dev)
            if t.commit():
                break

        # wait until device removal completes
        self.testDeviceComplete()
        self.testvifsComplete()

    def getDeviceController(self, name):
        if name not in controllerClasses:
            raise XendError("unknown device type: " + str(name))

        return controllerClasses[name](self)

    def getDiskConfigurations(self):
        vbd = self.getDeviceController('vbd')
        tap = self.getDeviceController('tap')
        ids = map(int, xstransact.List(vbd.frontendRoot()))

        result = []
        for disk in ids:
            if disk in tap.deviceIDs():
                config = tap.configuration(disk)
            else:
                config = vbd.configuration(disk)
            result.append(config)
        return result


    ## public:

    def sxpr(self):
        sxpr = ['domain',
                ['domid',   self.domid]]

        for e in ROUNDTRIPPING_CONFIG_ENTRIES:
            if self.infoIsSet(e[0]):
                sxpr.append([e[0], self.info[e[0]]])
        
        if self.infoIsSet('image'):
            sxpr.append(['image', self.info['image']])

        if self.infoIsSet('cpus'):
            sxpr.append(['cpus', self.info['cpus']])
 
        if self.infoIsSet('security'):
            sxpr.append(['security', self.info['security']])

        disks_done = 0
        for cls in controllerClasses:
            if cls in ('vbd', 'tap'):
                if not disks_done:
                    disks_done = 1
                    sxpr.extend(map(lambda x: ['device', x],
                                    self.getDiskConfigurations()))
                continue

            for config in self.getDeviceConfigurations(cls):
                sxpr.append(['device', config])

        def stateChar(name):
            if name in self.info:
                if self.info[name]:
                    return name[0]
                else:
                    return '-'
            else:
                return '?'

        state = reduce(
            lambda x, y: x + y,
            map(stateChar,
                ['running', 'blocked', 'paused', 'shutdown', 'crashed',
                 'dying']))

        sxpr.append(['state', state])
        if self.infoIsSet('shutdown'):
            reason = shutdown_reason(self.info['shutdown_reason'])
            sxpr.append(['shutdown_reason', reason])
        if self.infoIsSet('cpu_time'):
            sxpr.append(['cpu_time', self.info['cpu_time']/1e9])
        sxpr.append(['online_vcpus', self.info['online_vcpus']])
            
        if self.infoIsSet('start_time'):
            up_time =  time.time() - self.info['start_time']
            sxpr.append(['up_time', str(up_time) ])
            sxpr.append(['start_time', str(self.info['start_time']) ])

        if self.store_mfn:
            sxpr.append(['store_mfn', self.store_mfn])
        if self.console_mfn:
            sxpr.append(['console_mfn', self.console_mfn])

        return sxpr


    def getVCPUInfo(self):
        try:
            # We include the domain name and ID, to help xm.
            sxpr = ['domain',
                    ['domid',      self.domid],
                    ['name',       self.info['name']],
                    ['vcpu_count', self.info['online_vcpus']]]

            for i in range(0, self.info['max_vcpu_id']+1):
                info = xc.vcpu_getinfo(self.domid, i)

                sxpr.append(['vcpu',
                             ['number',   i],
                             ['online',   info['online']],
                             ['blocked',  info['blocked']],
                             ['running',  info['running']],
                             ['cpu_time', info['cpu_time'] / 1e9],
                             ['cpu',      info['cpu']],
                             ['cpumap',   info['cpumap']]])

            return sxpr

        except RuntimeError, exn:
            raise XendError(str(exn))
                      

    ## private:

    def check_name(self, name):
        """Check if a vm name is valid. Valid names contain alphabetic characters,
        digits, or characters in '_-.:/+'.
        The same name cannot be used for more than one vm at the same time.

        @param name: name
        @raise: VmError if invalid
        """
        if name is None or name == '':
            raise VmError('missing vm name')
        for c in name:
            if c in string.digits: continue
            if c in '_-.:/+': continue
            if c in string.ascii_letters: continue
            raise VmError('invalid vm name')

        dominfo = domain_by_name(name)
        if not dominfo:
            return
        if self.domid is None:
            raise VmError("VM name '%s' already in use by domain %d" %
                          (name, dominfo.domid))
        if dominfo.domid != self.domid:
            raise VmError("VM name '%s' is used in both domains %d and %d" %
                          (name, self.domid, dominfo.domid))


    def is_hvm(self):
        image = self.info['image']
        if sxp.name(image) == 'hvm':
            return 1
        else:
            return 0

    def construct(self):
        """Construct the domain.

        @raise: VmError on error
        """

        log.debug('XendDomainInfo.construct: %s',
                  self.domid)
        hvm = self.is_hvm()

        if hvm:
            info = xc.xeninfo()
            if 'hvm' not in info['xen_caps']:
                raise VmError("HVM guest support is unavailable: is VT/AMD-V "
                              "supported by your CPU and enabled in your "
                              "BIOS?")

        # Hack to pre-reserve some memory for initial domain creation.
        # There is an implicit memory overhead for any domain creation. This
        # overhead is greater for some types of domain than others. For
        # example, an x86 HVM domain will have a default shadow-pagetable
        # allocation of 1MB. We free up 4MB here to be on the safe side.
        # 2MB memory allocation was not enough in some cases, so it's 4MB now
        balloon.free(4*1024) # 4MB should be plenty

        self.domid = xc.domain_create(
            domid = 0, ssidref = security.get_security_info(self.info, 'ssidref'),
            handle = uuid.fromString(self.info['uuid']),
            hvm = hvm)

        if self.domid < 0:
            raise VmError('Creating domain failed: name=%s' %
                          self.info['name'])

        self.dompath = GetDomainPath(self.domid)

        self.recreateDom()

        # Set timer configration of domain
        timer_mode = self.info['timer_mode']
        if hvm:
            if timer_mode is None:
                timer_mode = 1
            xc.hvm_set_param(self.domid, HVM_PARAM_TIMER_MODE, long(timer_mode))

        # Set maximum number of vcpus in domain
        xc.domain_max_vcpus(self.domid, int(self.info['vcpus']))

        # Test whether the devices can be assigned.
        for (n,c) in self.info['device']:
            if not n or not c or not(n in ["pci"]):
                continue
            for i in c[1:]:
                if i[0] != 'dev':
                    continue
                seg = int(sxp.child_value(i, "domain"), 16)
                bus = int(sxp.child_value(i, "bus"), 16)
                dev = int(sxp.child_value(i, "slot"), 16)
                func = int(sxp.child_value(i, "func"), 16)

                pci_str = '%04x:%02x:%02x.%x' % (seg, bus, dev, func)
                from xen.xend.server.pciif import PciDevice
                try:
                    pci_device = PciDevice(seg, bus, dev, func)
                except Exception, e:
                    raise VmError("pci: failed to locate device and "+
                            "parse it's resources - "+str(e))
                if pci_device.driver!='pciback':
                    raise VmError(("pci: PCI Backend does not own device "+ \
                            "%s\n"+ \
                            "See the pciback.hide kernel "+ \
                            "command-line parameter or\n"+ \
                            "bind your slot/device to the PCI backend using sysfs" \
                            )%(pci_device.name))

                # Checking for hvm guest
                if self.is_hvm():
                    bdf = xc.test_assign_device(self.domid, seg, bus, dev, func)
                    if bdf != 0:
                        if bdf == -1:
                            raise VmError("failed to assign device: maybe the platform"
                                          " doesn't support VT-d, or VT-d isn't enabled"
                                          " properly?")
                        bus = (bdf >> 16) & 0xff
                        devfn = (bdf >> 8) & 0xff
                        dev = (devfn >> 3) & 0x1f
                        func = devfn & 0x7
                        raise VmError("failed to assign device(%x:%x.%x): maybe it has"
                                      " already been assigned to other domain, or maybe"
                                      " it doesn't exist." % (bus, dev, func))

                # Checking for both pv guests and hvm guests
                if pci_str in get_all_assigned_pci_devices():
                    raise VmError("failed to assign device %s that has"
                                  " already been assigned to other domain." % pci_str)

        # Check for cpu_{cap|weight} validity for credit scheduler if used
        if xen.lowlevel.xc.xc().sched_id_get() == xen.lowlevel.xc.XEN_SCHEDULER_CREDIT:
            cap = self.getCap()
            weight = self.getWeight()

            assert type(weight) == int
            assert type(cap) == int

            if weight < 1 or weight > 65535:
                raise VmError("Cpu weight out of range, valid values are within range from 1 to 65535")

            if cap < 0 or cap > self.getVCpuCount() * 100:
                raise VmError("Cpu cap out of range, valid range is from 0 to %s for specified number of vcpus" %
                              (self.getVCpuCount() * 100)) 

            xc.sched_credit_domain_set(self.domid, weight, cap)

    def introduceDomain(self):
        assert self.domid is not None
        assert self.store_mfn is not None
        assert self.store_port is not None

        try:
            IntroduceDomain(self.domid, self.store_mfn, self.store_port)
        except RuntimeError, exn:
            raise XendError(str(exn))

    def _setCPUAffinity(self):
        """ Repin domain vcpus if a restricted cpus list is provided
        """

        if self.info['cpus'] is not None and len(self.info['cpus']) > 0:
            for v in range(0, self.info['max_vcpu_id']+1):
                xc.vcpu_setaffinity(self.domid, v, self.info['cpus'])
        else:
            def find_relaxed_node(node_list):
                import sys
                nr_nodes = info['nr_nodes']
                if node_list is None:
                    node_list = range(0, nr_nodes)
                nodeload = [0]
                nodeload = nodeload * nr_nodes
                from xen.xend import XendDomain
                doms = XendDomain.instance().list()
                for dom in filter (lambda d: d.domid != self.domid, doms):
                    cpuinfo = dom.getVCPUInfo()
                    for vcpu in sxp.children(cpuinfo, 'vcpu'):
                        if sxp.child_value(vcpu, 'online') == 0: continue
                        cpumap = list(sxp.child_value(vcpu,'cpumap'))
                        for i in range(0, nr_nodes):
                            node_cpumask = info['node_to_cpu'][i]
                            for j in node_cpumask:
                                if j in cpumap:
                                    nodeload[i] += 1
                                    break
                for i in range(0, nr_nodes):
                    if len(info['node_to_cpu'][i]) > 0 and i in node_list:
                        nodeload[i] = int(nodeload[i] * 16 / len(info['node_to_cpu'][i]))
                    else:
                        nodeload[i] = sys.maxint
                return map(lambda x: x[0], sorted(enumerate(nodeload), key=lambda x:x[1]))

            info = xc.physinfo()
            if info['nr_nodes'] > 1:
                candidate_node_list = range(0, info['nr_nodes'])
                best_node = find_relaxed_node(candidate_node_list)[0]
                cpumask = info['node_to_cpu'][best_node]
                cores_per_node = info['nr_cpus'] / info['nr_nodes']
                nodes_required = (self.info['max_vcpu_id'] + 1 + cores_per_node - 1) / cores_per_node
                if nodes_required > 1:
                    log.debug("allocating %d NUMA nodes", nodes_required)
                    best_nodes = find_relaxed_node(filter(lambda x: x != best_node, range(0,info['nr_nodes'])))
                    for i in best_nodes[:nodes_required - 1]:
                        cpumask = cpumask + info['node_to_cpu'][i]
                for v in range(0, self.info['max_vcpu_id']+1):
                    xc.vcpu_setaffinity(self.domid, v, cpumask)

    def initDomain(self):
        log.debug('XendDomainInfo.initDomain: %s %s',
                  self.domid,
                  self.info['cpu_weight'])

        # if we have a boot loader but no image, then we need to set things
        # up by running the boot loader non-interactively
        if self.infoIsSet('bootloader') and not self.infoIsSet('image'):
            self.configure_bootloader()

        if not self.infoIsSet('image'):
            raise VmError('Missing image in configuration')

        try:
            self.image = image.create(self,
                                      self.info['image'],
                                      self.info['device'])

            localtime = self.info['localtime']
            if localtime is not None and localtime == 1:
                xc.domain_set_time_offset(self.domid)

            xc.domain_setcpuweight(self.domid, self.info['cpu_weight'])

            # Set credit scheduling only if credit scheduler is used
            if xen.lowlevel.xc.xc().sched_id_get() == xen.lowlevel.xc.XEN_SCHEDULER_CREDIT:
                xc.sched_credit_domain_set(self.getDomid(), self.getWeight(), self.getCap())

            # repin domain vcpus if a restricted cpus list is provided
            # this is done prior to memory allocation to aide in memory
            # distribution for NUMA systems.
            self._setCPUAffinity()

            # Use architecture- and image-specific calculations to determine
            # the various headrooms necessary, given the raw configured
            # values. maxmem, memory, and shadow are all in KiB.
            memory = self.image.getRequiredAvailableMemory(
                self.info['memory'] * 1024)
            maxmem = self.image.getRequiredAvailableMemory(
                self.info['maxmem'] * 1024)
            shadow = self.image.getRequiredShadowMemory(
                self.info['shadow_memory'] * 1024,
                self.info['maxmem'] * 1024)

            log.debug("_initDomain:shadow_memory=0x%x, maxmem=0x%x, memory=0x%x.", self.info['shadow_memory'], self.info['maxmem'], self.info['memory'],)
            # Round shadow up to a multiple of a MiB, as shadow_mem_control
            # takes MiB and we must not round down and end up under-providing.
            shadow = ((shadow + 1023) / 1024) * 1024

            # set memory limit
            xc.domain_setmaxmem(self.domid, maxmem)

            vtd_mem = 0
            # On VT-d enabled host, let's reserve extra memory for hvm guest.
            if self.is_hvm() and xc.test_assign_device(0, 0, 0, 0, 0) != -1:
                # Reserve 1 page per MiB of RAM for separate VT-d page table.
                vtd_mem = 4 * (maxmem / 1024)
                # Round vtd_mem up to a multiple of a MiB.
                vtd_mem = ((vtd_mem + 1023) / 1024) * 1024

            # Make sure there's enough RAM available for the domain
            balloon.free(memory + shadow + vtd_mem)

            # Set up the shadow memory
            shadow_cur = xc.shadow_mem_control(self.domid, shadow / 1024)
            self.info['shadow_memory'] = shadow_cur

            self.createChannels()

            channel_details = self.image.createImage()

            self.store_mfn = channel_details['store_mfn']
            if 'console_mfn' in channel_details:
                self.console_mfn = channel_details['console_mfn']
            if 'native_protocol' in channel_details:
                self.native_protocol = channel_details['native_protocol'];

            self.introduceDomain()

            self.createDevices()

            if self.info['bootloader']:
                self.image.cleanupBootloading()

            self.info['start_time'] = time.time()

        except RuntimeError, exn:
            if self.info['bootloader'] and self.image is not None:
                self.image.cleanupBootloading()
            raise VmError(str(exn))

    def get_list_difference(self, list1, list2):
        diff_list = []
        for item in list1:
            if not item in list2:
                diff_list.append(item)

        return diff_list

    ## public:
    def cleanupDomain(self):
        """Cleanup domain resources; release devices.  Idempotent.  Nothrow
        guarantee."""

        self.refresh_shutdown_lock.acquire()
        try:
            if (self.clean):
                return

            self.unwatchShutdown()

            self._releaseDevices()

            if self.image:
                try:
                    self.image.destroy()
                except:
                    log.exception(
                        "XendDomainInfo.cleanup: image.destroy() failed.")
                self.image = None

            try:
                self.removeDom()
            except:
                log.exception("Removing domain path failed.")

            try:
                if not self.info['name'].startswith(ZOMBIE_PREFIX):
                    name = ZOMBIE_PREFIX + self.info['name']
                    self.storeVm('name', name)
                    self.info['name'] = name
            except:
                log.exception("Renaming Zombie failed.")

            # Release remaining PVFB devices for domains that are not
            # currently in use so we need to get currently active domains
            activeIds = []
            val = XendDomain.instance().domains.values()
            for v in val:
                id = v.domid
                if (id > 0):
                    try:
                        index = activeIds.index(id)
                    except:
                        activeIds.append(id)

            # Then we get list of all set PVFB devices
            idPVFB = []
            t = xstransact("/local/domain/0/backend/vfb")
            for dev in t.list():
                if int(dev) <= self.domid:
                    idPVFB.append( int(dev) )

            # Finally we get the difference to compile a list of
            # zombie IDs to be detached (not to cause xenstore leaks)
            idDel = self.get_list_difference( idPVFB, activeIds )

            # We also check entry in /vm/uuid to check if we have one
            uuidCreated = False
            t = xstransact("/vm/%s" % self.info['uuid'] )
            for dev in t.list():
                if dev == 'uuid':
                    uuidCreated = True

            log.debug("UUID Created: %s" % uuidCreated)
            log.debug("Devices to release: %s, domid = %d" % (idDel, self.domid) )

            # If there is no new uuid entry, release front-end
            if not uuidCreated:
                log.debug("Releasing PVFB front-end devices (uuid not created)...")
                for n in ['vfb', 'vkbd']:
                    t = xstransact("/vm/%s/device/%s" % (self.info['uuid'], n) )
                    for dev in t.list():
                        t.remove(dev)
                    t.commit()

            # Finally release backend devices for all IDs in the list
            log.debug("Releasing PVFB backend devices ...")
            for n in ['vfb', 'vkbd']:
                for id in idDel:
                    t = xstransact("/local/domain/0/backend/%s/%d" % (n, id) )
                    for dev in t.list():
                        t.remove(dev)
                    t.commit()


            self.state_set(STATE_DOM_SHUTDOWN)
            self.clean = True
        finally:
            self.refresh_shutdown_lock.release()


    def cleanupVm(self):
        """Cleanup VM resources.  Idempotent.  Nothrow guarantee."""

        self.unwatchVm()

        try:
            self.removeVm()
        except:
            log.exception("Removing VM path failed.")


    ## private:

    def unwatchVm(self):
        """Remove the watch on the VM path, if any.  Idempotent.  Nothrow
        guarantee."""

        try:
            try:
                if self.vmWatch:
                    self.vmWatch.unwatch()
            finally:
                self.vmWatch = None
        except:
            log.exception("Unwatching VM path failed.")


    def unwatchShutdown(self):
        """Remove the watch on the domain's control/shutdown node, if any.
        Idempotent.  Nothrow guarantee.  Expects to be protected by the
        refresh_shutdown_lock."""

        try:
            try:
                if self.shutdownWatch:
                    self.shutdownWatch.unwatch()
            finally:
                self.shutdownWatch = None
        except:
            log.exception("Unwatching control/shutdown failed.")


    ## public:

    def destroy(self):
        """Cleanup VM and destroy domain.  Nothrow guarantee."""

        log.debug("XendDomainInfo.destroy: domid=%s", self.domid)

        if self.dompath is not None:
            try:
                xc.domain_destroy_hook(self.domid)
                xc.domain_pause(self.domid)
                do_FLR(self.domid,  self.is_hvm())
                xc.domain_destroy(self.domid)
            except:
                log.exception("XendDomainInfo.destroy: xc.domain_destroy failed.")
            self.cleanupDomain()

        self.cleanupVm()

    ## private:

    def resumeDomain(self):
        log.debug("XendDomainInfo.resumeDomain(%s)", str(self.domid))

        if self.domid is None:
            return
        try:
            # could also fetch a parsed note from xenstore
            # XXX we're missing some of the SUSPEND_CANCEL stuff
            #fast = self.info.get_notes().get('SUSPEND_CANCEL') and 1 or 0
            fast = self.is_hvm() 
            if not fast:
                self._releaseDevices()
                log.debug("XendDomainInfo.resumeDomain: devices released")

                self._resetChannels()

                self.removeDom('control/shutdown')
                self.removeDom('device-misc/vif/nextDeviceID')

                self.createChannels()
                self.introduceDomain()
                self.storeDomDetails()

                self.createDevices()
                log.debug("XendDomainInfo.resumeDomain: devices created")

            xc.domain_resume(self.domid, fast)
            ResumeDomain(self.domid)
        except:
            log.exception("XendDomainInfo.resume: xc.domain_resume failed on domain %s." % (str(self.domid)))

    def createChannels(self):
        """Create the channels to the domain.
        """
        self.store_port = self.createChannel()
        self.console_port = self.createChannel()


    def createChannel(self):
        """Create an event channel to the domain.
        """
        try:
            return xc.evtchn_alloc_unbound(domid=self.domid, remote_dom=0)
        except:
            log.exception("Exception in alloc_unbound(%d)", self.domid)
            raise

    def _resetChannels(self):
        """Reset all event channels in the domain.
        """
        try:
            if self.domid != None:
                return xc.evtchn_reset(dom = self.domid)
        except:
            log.exception("Exception in evtcnh_reset(%s)", str(self.domid))
            raise


    ## public:

    def createDevices(self):
        """Create the devices for a vm.

        @raise: VmError for invalid devices
        """

        # vkbd devices *must* be before vfb devices, otherwise
        # there is a race condition when setting up devices
        # where the daemon spawned for the vfb may write stuff
        # into xenstore vkbd backend, before DevController has
        # setup permissions on the vkbd backend path. This race
        # results in domain creation failing with 'device already
        # connected' messages
        devs = []
        for d in self.info['device']:
            if d[0] == 'vkbd':
                devs.insert(0, d)
            else:
                devs.append(d)
        for (n, c) in devs:
            self.createDevice(n, c)

        if self.image:
            self.image.createDeviceModel()

        #if have pass-through devs, need the virtual pci slots info from qemu
        self.sync_pcidev_info()

    ## public:

    def checkLiveMigrateMemory(self):
        """ Make sure there's enough memory to migrate this domain """
        overhead_kb = 0
        if arch.type == "x86":
            # 1MB per vcpu plus 4Kib/Mib of RAM.  This is higher than 
            # the minimum that Xen would allocate if no value were given.
            overhead_kb = self.info['vcpus'] * 1024 + self.info['maxmem'] * 4
            overhead_kb = ((overhead_kb + 1023) / 1024) * 1024
            # The domain might already have some shadow memory
            overhead_kb -= xc.shadow_mem_control(self.domid) * 1024
        if overhead_kb > 0:
            balloon.free(overhead_kb)

    def testMigrateDevices(self, network, dst):
        """ Notify all device about intention of migration
        @raise: XendError for a device that cannot be migrated
        """
        for (n, c) in self.info['device']:
            rc = self.migrateDevice(n, c, network, dst, DEV_MIGRATE_TEST)
            if rc != 0:
                raise XendError("Device of type '%s' refuses migration." % n)

    def testDeviceComplete(self):
        """ For Block IO migration safety we must ensure that
        the device has shutdown correctly, i.e. all blocks are
        flushed to disk
        """
        start = time.time()
        while True:
            test = 0
            diff = time.time() - start
            vbds = self.getDeviceController('vbd').deviceIDs()
            taps = self.getDeviceController('tap').deviceIDs()
            for i in vbds + taps:
                test = 1
                log.info("Dev %s still active, looping...", i)
                time.sleep(0.1)
                
            if test == 0:
                break
            if diff >= MIGRATE_TIMEOUT:
                log.info("Dev still active but hit max loop timeout")
                break

    def migrateDevices(self, network, dst, step, domName=''):
        """Notify the devices about migration
        """
        ctr = 0
        try:
            for (n, c) in self.info['device']:
                self.migrateDevice(n, c, network, dst, step, domName)
                ctr = ctr + 1
        except:
            for (n, c) in self.info['device']:
                if ctr == 0:
                    step = step - 1
                ctr = ctr - 1
                self.recoverMigrateDevice(n, c, network, dst, step, domName)
            raise

    def migrateDevice(self, deviceClass, deviceConfig, network, dst,
                      step, domName=''):
        return self.getDeviceController(deviceClass).migrate(deviceConfig,
                                        network, dst, step, domName)

    def recoverMigrateDevice(self, deviceClass, deviceConfig, network,
                             dst, step, domName=''):
        return self.getDeviceController(deviceClass).recover_migrate(
                     deviceConfig, network, dst, step, domName)

    def waitForDevices(self):
        """Wait for this domain's configured devices to connect.

        @raise: VmError if any device fails to initialise.
        """
        for c in controllerClasses:
            self.waitForDevices_(c)

    def sync_pcidev_info(self):

        if not self.is_hvm():
            return

        devid = '0'
        dev_info = self._getDeviceInfo_pci()
        if dev_info is None:
            return

        dev_list = sxp.children(dev_info, 'dev')

        # get the virtual slot info from xenstore
        count = 0
        vslots = None
        while vslots is None and count < 20:
            vslots = xstransact.Read("/local/domain/0/backend/pci/%u/%s/vslots"
                              % (self.getDomid(), devid))
            time.sleep(0.1)
            count += 1
        if vslots is None:
            log.error("Device model didn't tell the vslots for PCI device")
            return

        #delete last delim
        if vslots[-1] == ";":
            vslots = vslots[:-1]

        slot_list = vslots.split(';')
        if len(slot_list) != len(dev_list):
            log.error("Device model's pci dev num dismatch")
            return

        #update the vslot info
        count = 0;
        for x in dev_info:
            if sxp.name(x) != 'dev':
                continue
            vslt_info = ['vslt', slot_list[count]]
            x.append(vslt_info)
            count += 1


    def pci_device_create(self, dev_sxp):
        log.debug("XendDomainInfo.pci_device_create: %s" % (dev_sxp))

        if not self.is_hvm():
            raise VmError("only HVM guest support pci attach")

        #all the PCI devs share one conf node
        devid = '0'

        dev_type = sxp.name(dev_sxp)
        dev_config = pci_convert_sxp_to_dict(dev_sxp)
        new_pci = dev_config['devs'][0]

        new_seg  = int(new_pci['domain'], 16)
        new_bus  = int(new_pci['bus'], 16)
        new_dev  = int(new_pci['slot'], 16)
        new_func = int(new_pci['func'], 16)
        if new_pci.has_key('vslt'):
            new_vslt = int(new_pci['vslt'], 16)
        else:
            new_vslt = 0x0
            new_pci['vslt'] = '0x0'

        #check conflict before trigger hotplug event
        dev_info = self._getDeviceInfo_pci()
        if dev_info is not None:
            for x in dev_info:
                if sxp.name(x) != 'dev':
                    continue
                seg  = int(sxp.child_value(x, "domain"), 16)
                bus  = int(sxp.child_value(x, "bus"), 16)
                dev  = int(sxp.child_value(x, "slot"), 16)
                func = int(sxp.child_value(x, "func"), 16)
                vslt = int(sxp.child_value(x, "vslt"), 16)

                if (vslt == new_vslt and vslt != 0 ):
                    raise VmError("vslot %s already have a device." % (new_pci['vslt']))

                if (seg  == new_seg and
                   bus  == new_bus and
                   dev  == new_dev and
                   func == new_func ):
                    raise VmError("device is already inserted")

        bdf = xc.test_assign_device(self.domid, new_seg, new_bus, new_dev, new_func)
        pci_str = '%04x:%02x:%02x.%x' % (new_seg, new_bus, new_dev, new_func)

        if bdf != 0:
            if bdf == -1:
                raise VmError("failed to assign device: maybe the platform"
                              " doesn't support VT-d, or VT-d isn't enabled"
                              " properly?")
            bus = (bdf >> 16) & 0xff
            devfn = (bdf >> 8) & 0xff
            dev = (devfn >> 3) & 0x1f
            func = devfn & 0x7
            raise VmError("failed to hot insert device(%x:%x.%x): maybe VT-d is "
                          "not enabled, or the device is not exist, or it "
                          "has already been assigned to other domain"
                          % (bus, dev, func))

        if pci_str in get_all_assigned_pci_devices():
            raise VmError("failed to assign device %s that has"
                          " already been assigned to other domain." % pci_str)

        # Here, we duplicate some checkings logic: in some cases, we mustn't
        # allow a device to be hot-plugged into an HVM guest.
        # We must make the checkings before sending the command 'pci-ins' to
        # ioemu.

        # Test whether the device is owned by pciback. For instance, we can't
        # hotplug a device being used by Dom0 itself to an HVM guest.
        from xen.xend.server.pciif import PciDevice, parse_pci_name
        try:
            pci_device = PciDevice(new_seg, new_bus, new_dev, new_func)
        except Exception, e:
            raise VmError("pci: failed to locate device and "+
                    "parse it's resources - "+str(e))
        if pci_device.driver!='pciback':
            raise VmError(("pci: PCI Backend does not own device "+ \
                    "%s\n"+ \
                    "See the pciback.hide kernel "+ \
                    "command-line parameter or\n"+ \
                    "bind your slot/device to the PCI backend using sysfs" \
                    )%(pci_device.name))

        # Check non-page-aligned MMIO BAR.
        if pci_device.has_non_page_aligned_bar:
            raise VmError("pci: %s: non-page-aligned MMIO BAR found." % \
                pci_device.name)

        # Check if there is intermediate PCIe switch bewteen the device and
        # Root Complex.
        if pci_device.is_behind_switch_lacking_acs() and xroot.get_pci_strict_checking():
            err_msg = 'pci: to avoid potential security issue, %s is not'+\
                    ' allowed to be assigned to guest since it is behind'+\
                    ' PCIe switch that does not support or enable ACS.'
            raise VmError(err_msg % pci_device.name)

        # Check the co-assignment.
        # To pci-attach a device D to domN, we should ensure each of D's
        # co-assignment devices hasn't been assigned, or has been assigned to
        # domN.
        coassignment_list = pci_device.find_coassigned_devices()
        pci_device.devs_check_driver(coassignment_list)
        assigned_pci_device_str_list = get_assigned_pci_devices(self.domid)
        for pci_str in coassignment_list:
            if not pci_str in get_all_assigned_pci_devices():
                continue
            if not pci_str in assigned_pci_device_str_list:
                raise VmError(("pci: failed to pci-attach %s to dom%d" + \
                    " because one of its co-assignment device %s has been" + \
                    " assigned to other domain." \
                    )% (pci_device.name, self.domid, pci_str))

        bdf_str = "%s:%s:%s.%s@%s" % (new_pci['domain'],
                new_pci['bus'],
                new_pci['slot'],
                new_pci['func'],
                new_pci['vslt'])
        self.image.signalDeviceModel('pci-ins', 'pci-inserted', bdf_str)

        # update the virtual pci slot
        vslt = xstransact.Read("/local/domain/0/device-model/%i/parameter"
                          % self.getDomid())
        new_pci['vslt'] = vslt

        dev_cfg = ['dev',
              ['domain', new_pci['domain']],
              ['bus',    new_pci['bus']],
              ['slot',   new_pci['slot']],
              ['func',   new_pci['func']],
              ['vslt',   new_pci['vslt']]
            ]

        if dev_info is None:
            # create a new one from scrach
            dev_cfg_sxp = [dev_type, dev_cfg]
            try:
                devid = self.createDevice(dev_type, dev_cfg_sxp)
                self.waitForDevice(dev_type, devid)
                self.info['device'].append((dev_type, dev_cfg_sxp))
                #return self.getDeviceController(dev_type).sxpr(devid)
            except VmError, ex:
                raise ex

        else:
            # update the pci config to add the new dev
            dev_info.append(dev_cfg)
            self.reconfigureDevice('pci', devid, dev_info)

        return self.getDeviceController(dev_type).sxpr(devid)



    def device_create(self, dev_config):
        """Create a new device.

        @param dev_config: device configuration
        """
        dev_type = sxp.name(dev_config)

        if dev_type == 'vif':
            for x in dev_config:
                if x != 'vif' and x[0] == 'mac':
                    if not re.match('^([0-9a-f]{2}:){5}[0-9a-f]{2}$', x[1], re.I):
                        log.error("Virtual network interface creation error - invalid MAC Address entered: %s", x[1])
                        raise VmError("Cannot create a new virtual network interface - MAC address is not valid!");
        elif dev_type == 'pci':
            # NOTE: for pci, the dev_config here is actually the sxp
            # style list, something like:
            # ['pci', ['dev', ['domain', '0x0000'], ['bus', '0x08'], ['slot', '0x00'], ['func', '0x0']]]
            # this will be converted as appropriate in pci_device_create
            rc = self.pci_device_create(dev_config)
            return rc

        devid = self.createDevice(dev_type, dev_config)
        self.waitForDevice(dev_type, devid)
        self.info['device'].append((dev_type, dev_config))
        return self.getDeviceController(dev_type).sxpr(devid)


    def device_configure(self, dev_config, devid=None):
        """Configure an existing device.
        @param dev_config: device configuration
        """
        deviceClass = sxp.name(dev_config)
        # HACK: upstream uses device_configure to implement pci_detach.
        # In order to let external HTTP clients (like libvirt) be compatible
        # with that, we add a custom rule here to do detach
        if deviceClass == 'pci' and self.is_hvm():
            if sxp.child_value(dev_config, 'state') == 'Closing':
                return self.destroyPCIDevice(dev_config)
        self.reconfigureDevice(deviceClass, devid, dev_config)


    def pause(self):
        xc.domain_pause(self.domid)


    def unpause(self):
        xc.domain_unpause(self.domid)


    ## private:

    def restart(self, rename = False):
        """Restart the domain after it has exited.

        @param rename True if the old domain is to be renamed and preserved,
        False if it is to be destroyed.
        """

        self.configure_bootloader()
        config = self.sxpr()

        if self.infoIsSet('cpus') and len(self.info['cpus']) != 0:
            config.append(['cpus', reduce(lambda x, y: str(x) + "," + str(y),
                                          self.info['cpus'])])

        if self.readVm(RESTART_IN_PROGRESS):
            log.error('Xend failed during restart of domain %d.  '
                      'Refusing to restart to avoid loops.',
                      self.domid)
            self.destroy()
            return

        self.writeVm(RESTART_IN_PROGRESS, 'True')

        now = time.time()
        rst = self.readVm('xend/previous_restart_time')
        if rst:
            rst = float(rst)
            timeout = now - rst
            if timeout < MINIMUM_RESTART_TIME:
                log.error(
                    'VM %s restarting too fast (%f seconds since the last '
                    'restart).  Refusing to restart to avoid loops.',
                    self.info['name'], timeout)
                self.destroy()
                return

        self.writeVm('xend/previous_restart_time', str(now))

        prev_vm_xend = self.listRecursiveVm('xend')
        try:
            if rename:
                self.preserveForRestart()
            else:
                self.unwatchVm()
                self.destroy()

            # new_dom's VM will be the same as this domain's VM, except where
            # the rename flag has instructed us to call preserveForRestart.
            # In that case, it is important that we remove the
            # RESTART_IN_PROGRESS node from the new domain, not the old one,
            # once the new one is available.

            new_dom = None
            try:
                new_dom = XendDomain.instance().domain_create(config)
                for x in prev_vm_xend[0][1]:
                    new_dom.writeVm('xend/%s' % x[0], x[1])
                new_dom.waitForDevices()
                new_dom.unpause()
                rst_cnt = new_dom.readVm('xend/restart_count')
                rst_cnt = int(rst_cnt) + 1
                new_dom.writeVm('xend/restart_count', str(rst_cnt))
                new_dom.removeVm(RESTART_IN_PROGRESS)
            except:
                if new_dom:
                    new_dom.removeVm(RESTART_IN_PROGRESS)
                    new_dom.destroy()
                else:
                    self.removeVm(RESTART_IN_PROGRESS)
                raise
        except:
            log.exception('Failed to restart domain %d.', self.domid)


    def preserveForRestart(self):
        """Preserve a domain that has been shut down, by giving it a new UUID,
        cloning the VM details, and giving it a new name.  This allows us to
        keep this domain for debugging, but restart a new one in its place
        preserving the restart semantics (name and UUID preserved).
        """
        
        new_name = self.generateUniqueName()
        new_uuid = uuid.toString(uuid.create())
        log.info("Renaming dead domain %s (%d, %s) to %s (%s).",
                 self.info['name'], self.domid, self.info['uuid'],
                 new_name, new_uuid)
        self.unwatchVm()
        self._releaseDevices()
        # Remove existing vm node in xenstore
        self.removeVm()
        self.info['name'] = new_name
        self.info['uuid'] = new_uuid
        self.vmpath = XendDomain.VMROOT + new_uuid
        self.storeVmDetails()
        self.preserve()


    def preserve(self):
        log.info("Preserving dead domain %s (%d).", self.info['name'],
                 self.domid)
        self.unwatchVm()
        self.storeDom('xend/shutdown_completed', 'True')
        self.state_set(STATE_DOM_SHUTDOWN)


    # private:

    def generateUniqueName(self):
        n = 1
        while True:
            name = "%s-%d" % (self.info['name'], n)
            try:
                self.check_name(name)
                return name
            except VmError:
                n += 1


    def configure_bootloader(self):
        """Run the bootloader if we're configured to do so."""
        if not self.info['bootloader']:
            return
        blcfg = None
        # FIXME: this assumes that we want to use the first disk device
        for (n,c) in self.info['device']:
            if not n or not c or not(n in ["vbd", "tap"]):
                continue
            disk = sxp.child_value(c, "uname")
            if disk is None:
                continue
            fn = blkdev_uname_to_file(disk)
            blcfg = bootloader(self.info['bootloader'], fn, 1,
                               self.info['bootloader_args'],
                               self.info['image'])
            break
        if blcfg is None:
            msg = "Had a bootloader specified, but can't find disk"
            log.error(msg)
            raise VmError(msg)
        self.info['image'] = blcfg


    def send_sysrq(self, key):
        asserts.isCharConvertible(key)

        self.storeDom("control/sysrq", '%c' % key)


    def infoIsSet(self, name):
        return name in self.info and self.info[name] is not None


#============================================================================
# Register device controllers and their device config types.

"""A map from device-class names to the subclass of DevController that
implements the device control specific to that device-class."""
controllerClasses = {}

def addControllerClass(device_class, cls):
    """Register a subclass of DevController to handle the named device-class.
    """
    cls.deviceClass = device_class
    controllerClasses[device_class] = cls


from xen.xend.server import blkif, netif, tpmif, pciif, iopif, irqif, usbif, vfbif
from xen.xend.server.BlktapController import BlktapController
addControllerClass('vbd',  blkif.BlkifController)
addControllerClass('vif',  netif.NetifController)
addControllerClass('vtpm', tpmif.TPMifController)
addControllerClass('pci',  pciif.PciController)
addControllerClass('ioports', iopif.IOPortsController)
addControllerClass('irq',  irqif.IRQController)
addControllerClass('usb',  usbif.UsbifController)
addControllerClass('tap',  BlktapController)
addControllerClass('vfb',  vfbif.VfbifController)
addControllerClass('vkbd',  vfbif.VkbdifController)
